/* ==========================================================================
   talk - A talk client for MS-DOS. Uses the WATTCP TCP/IP library.

   (C) 1993 by Michael Ringe, Institut fuer Theoretische Physik,
	       RWTH Aachen, Aachen, Germany
	       <michael@thphys.physik.rwth-aachen.de>

   This program is free software; see the file COPYING for details.
   ================================================================== */

#define DEBUG 0
#define VERSION "1.0"

#define A_LOCAL 0x0F		// Default screen attributes
#define A_REMOTE 0x0E
#define A_STATUS 0x1C
#define A_STATUSL 0x1F
#define A_STATUSR 0x1E
#define A_NORM 0x07

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include "..\include\tcp.h"

#define MY_DATA_PORT 1050
#define MY_CTL_PORT 1051
#define TALK_PORT 518
#define MSG_ID 1234568


// Global data

char
    *localuser = "PC-User",	// My user name
    *remoteuser = NULL,		// Remote user name
    *remotehost = NULL;		// Remote host name
longword remoteip = 0;		// Remote IP address
tcp_Socket ds;			// Data socket
char scrnmode = 0;		// Screen mode
int a_local = A_LOCAL,		// Screen attributes
    a_remote = A_REMOTE,
    a_status = A_STATUS,
    a_statusl = A_STATUSL,
    a_statusr = A_STATUSR;


// The following stuff is adapted from Linux's talkd and talk source

typedef struct {
	byte	 vers;
	byte	 type;
	byte	 answer;
	byte	 pad;
	longword id_num;
	struct	 sockaddr addr;
	short    pad1;		// because WATTCP's sockaddr is only 14 bytes
	struct	 sockaddr ctl_addr;
	short    pad2;		// see above
	long	 pid;
#define	NAME_SIZE	12
	char	 l_name[NAME_SIZE];
	char	 r_name[NAME_SIZE];
#define	TTY_SIZE	16
	char	 r_tty[TTY_SIZE];
} CTL_MSG;

typedef struct {
	byte	 vers;
	byte	 type;
	byte	 answer;
	byte	 pad;
	longword id_num;
	struct	 sockaddr addr;
	short    pad1;		// see above
} CTL_RESPONSE;

#define TALK_VERSION	1

#define LEAVE_INVITE	0
#define LOOK_UP		1
#define DELETE		2
#define ANNOUNCE	3

#define SUCCESS		0
#define NOT_HERE	1
#define FAILED		2
#define MACHINE_UNKNOWN	3
#define PERMISSION_DENIED 4
#define UNKNOWN_REQUEST	5
#define	BADVERSION	6
#define	BADADDR		7
#define	BADCTLADDR	8
#define MAXANSWER	8

static char *answermsg[MAXANSWER+1] = {
	"",
	"Your party is not logged on",
	"Target machine is too confused to talk to us",
	"Target machine does not recognize us",
	"Your party is refusing messages",
	"Target machine can not handle remote talk",
	"Target machine indicates protocol mismatch",
	"Target machine indicates protocol botch (addr)",
	"Target machine indicates protocol botch (ctl_addr)"
};


// --------------------------------------------------------------------------
// Some utility functions:
// ip2a() - Convert long to ascii in IP dotted decimal notation
// mystrncpy() - Safe strncpy
// setsockaddr() - Set socketaddr fields in network byte order
// --------------------------------------------------------------------------

static char *ip2a(longword l)

{
    static char x[20];

    sprintf(x,"%d.%d.%d.%d",
     ((unsigned char *)&l)[3],
     ((unsigned char *)&l)[2],
     ((unsigned char *)&l)[1],
     ((unsigned char *)&l)[0]
     );
    return x;
}

static char *mystrncpy(char *dest, char *src, int len)

{
    strncpy(dest,src,len);
    dest[len-1] = 0;
    return dest;
}

static void setsockaddr(struct sockaddr *s, word port, longword ip)

{
    s->s_type = 0x0200;		/* AF_INET */
    s->s_port = intel16(port);
    s->s_ip = intel(ip);
    memset(s->s_spares,0,6);
}


// --------------------------------------------------------------------------
// Initialization stuff
// --------------------------------------------------------------------------

static void my_usr_init(char *key, char *val)

{
    static char usrname[30];
    if (!stricmp(key,"talk_localname"))
    {
	strncpy(usrname,val,30);
	localuser = usrname;
    }
    else if (!stricmp(key,"talk_screenmode"))
    {
	if (!stricmp(val,"split"))
	    scrnmode = 1;
    }
    else if (!stricmp(key,"talk_colors"))
    {
	sscanf(val,"%2x%2x%2x%2x%2x",&a_local,&a_remote,&a_status,
		&a_statusl,&a_statusr);
    }
}

static int init(int argc, char *argv[])

{
    char *c;
    int x = wherex(), y = wherey();

    window(1,1,80,25);
    textattr(A_NORM);
    gotoxy(x,y);
    cputs("Talk version " VERSION ", written by Michael Ringe\n\r");

    /* Get user and host from command line */
    if (argc > 2 || (argc == 2 && (c = strchr(argv[1],'@')) == NULL))
    {	cputs("Usage: talk [ user@host ]\n\r");
	return 1;
    }
    if (argc == 2)
    {
	*c++ = 0;
	remoteuser = argv[1];
	remotehost = c;
    }
#if DEBUG
    cprintf("Local user: %s\n\rRemote user: %s\n\rRemote host: %s\n\r",
	localuser,remoteuser,remotehost);
#endif

    /* Look up remote IP address */
    usr_init = my_usr_init;
    sock_init();

    if (argc != 2) return 0;
    if ( !(remoteip = resolve(remotehost)))
    {
	cprintf("Error: Cannot resolve '%s'\n\r",remotehost);
	return 1;
    }
#if DEBUG
    cprintf("Remote IP: %s\n\r",ip2a(remoteip));
#endif
    return 0;
}




// --------------------------------------------------------------------------
// wait_invite() - Wait for a remote invitation
// --------------------------------------------------------------------------

static int wait_invite(void)

{
    udp_Socket daemon;
    CTL_MSG ctl;
    CTL_RESPONSE ans;
    word status;
    static char magic[3] = { 0x7f,0x15,0x17 };
    static char magic2[3] = { 0,0,0 };
    int success = 0;

    if (!udp_open(&daemon,TALK_PORT,0,0,NULL))
	goto sock_err;
#if DEBUG
    cprintf("talkd socket opened\n\r");
#endif

    if (remoteip == 0)
	cputs("Waiting for incoming talk request");
    else
	cputs("Waiting for your party to respond");
    cputs("... press ESC to quit\n\r");

    while (!kbhit() || getch() != 27)
    {
	tcp_tick(&daemon);
	ctl.vers = TALK_VERSION + 1;
	if (sock_fastread(&daemon,(byte *)&ctl,sizeof(ctl)) != sizeof(ctl))
	    continue;
	if (ctl.vers != TALK_VERSION || ctl.type != LOOK_UP ) continue;
#if DEBUG
	cprintf("talkd: received lookup request, addr = %s:%u,  ",
		ip2a(intel(ctl.addr.s_ip)),intel16(ctl.addr.s_port));
	cprintf("ctl_addr=%s:%u name=%s\n\r",
		ip2a(intel(ctl.ctl_addr.s_ip)),intel16(ctl.ctl_addr.s_port),
		ctl.r_name);
#endif
	success = 1;
	break;
    }

    if (!success) return 1;
    ans.vers = TALK_VERSION;
    ans.type = LOOK_UP;
    ans.answer = SUCCESS;
    ans.id_num = MSG_ID;
    setsockaddr(&ans.addr,MY_DATA_PORT,my_ip_addr);
    sock_write(&daemon,(char *)&ans,sizeof(ans));
    sock_close(&daemon);

#if DEBUG
    cputs("Replied with SUCCESS.\n\r");
#endif

    tcp_listen( &ds, MY_DATA_PORT, 0, 0, NULL, 0 );
    sock_wait_established( &ds, 0, NULL, &status);
    sock_wait_input( &ds, 0, NULL, &status);
    sock_read(&ds,magic2,3);
    sock_write(&ds,magic,3);

    if (remoteip == 0)
    {
	static char remoteipstr[16], remoteuserstr[NAME_SIZE],
		localuserstr[NAME_SIZE];
	struct sockaddr sa;
	int l = sizeof(sa);

	getpeername(&ds,&sa,&l);
	strcpy(remoteipstr,ip2a(sa.s_ip));
	remotehost = remoteipstr;
	mystrncpy(remoteuserstr,ctl.l_name,NAME_SIZE);
	remoteuser = remoteuserstr;
	mystrncpy(localuserstr,ctl.r_name,NAME_SIZE);
	localuser = localuserstr;
    }
    return 0;

sock_err:
    return 1;
}



// --------------------------------------------------------------------------
// invite() - Announce an invitation to the remote talkd.
// --------------------------------------------------------------------------

static int invite(void)

{
    udp_Socket cs;
    CTL_MSG ctl;
    CTL_RESPONSE ans;
    word status;
    int count;
    static char magic[3] = { 0x7f,0x15,0x17 };
    static char magic2[3] = { 0,0,0 };

    // Open control socket
    // -------------------
    if (!udp_open(&cs,MY_CTL_PORT,remoteip,TALK_PORT,NULL))
	goto sock_err;
    sock_wait_established(&cs,sock_delay,NULL,&status);
#if DEBUG
    cprintf("Control socket opened\n\r");
#endif


    // Check for invitation on remote machine
    // --------------------------------------
#if DEBUG
    cprintf("Lookup on remote machine\n\r");
#endif
    ctl.vers = TALK_VERSION;
    ctl.type = LOOK_UP;
    ctl.id_num = MSG_ID;
    setsockaddr(&ctl.addr,0,0);
    setsockaddr(&ctl.ctl_addr,MY_CTL_PORT,my_ip_addr);
    mystrncpy(ctl.l_name,localuser,NAME_SIZE);
    mystrncpy(ctl.r_name,remoteuser,NAME_SIZE);
    mystrncpy(ctl.r_tty,"",TTY_SIZE);

    for (count = 3; count > 0; --count)
    {
#if DEBUG
	putch('.');
#endif
	sock_write(&cs,(byte *)&ctl,sizeof(ctl));
	if (_ip_delay1(&cs,3,NULL,&status)) continue;
	if (sock_read(&cs,(byte *)&ans,sizeof(ans)) != sizeof(ans)) continue;
	if (ans.vers != TALK_VERSION || ans.type != LOOK_UP) continue;
#if DEBUG
	cprintf("Remote talkd response: Version=%d, Type=%d, Answer=%d\n\r",
	    ans.vers,ans.type,ans.answer);
	cprintf("addr=%s:%u\n\r",ip2a(ans.addr.s_ip),intel16(ans.addr.s_port));
#endif
	if (ans.answer == SUCCESS)
	{
	    tcp_open(&ds,MY_DATA_PORT,intel(ans.addr.s_ip),intel16(ans.addr.s_port),NULL);
	    sock_wait_established( &ds, 0, NULL, &status);
	    sock_write(&ds,magic,3);
	    sock_wait_input( &ds, 0, NULL, &status);
	    sock_read(&ds,magic2,3);
	    return 0;
	}
	else
	    break;
    }

    // Announce invitation
    // -------------------
#if DEBUG
    cprintf("Announcing invitation to remote talkd\n\r");
#endif
    ctl.type = ANNOUNCE;
    setsockaddr(&ctl.addr,MY_DATA_PORT,my_ip_addr);

    for (count = 3; count > 0; --count)
    {
	int a;
#if DEBUG
	putch('.');
#endif
	sock_write(&cs,(byte *)&ctl,sizeof(ctl));
	if (_ip_delay1(&cs,sock_delay,NULL,&status)) continue;
	if (sock_read(&cs,(byte *)&ans,sizeof(ans)) != sizeof(ans)) continue;
	a = ans.answer;
#if DEBUG
	cprintf("Remote talkd response: Version=%d, Type=%d, Answer=%d\n\r",
	    ans.vers,ans.type,ans.answer);
#endif
	if (ans.vers != TALK_VERSION || ans.type != ANNOUNCE) continue;

#if DEBUG
	cprintf("Valid answer %d\n\r",a);
#endif
	if (a == SUCCESS)
	{
	    if (wait_invite())
	    {
#if DEBUG
		cputs("Deleting invitation.\n\r");
#endif
		ctl.type = DELETE;
		sock_write(&cs,(byte *)&ctl,sizeof(ctl));
		return 1;
	    }
	    else
		return 0;
	}
	else if (a < MAXANSWER)
	{
	    cputs(answermsg[a]);
	    cputs("\n\r");
	}
	else
	    cprintf("Unknown error %d from remote talkd\n\r",a);
	return 1;
    }

sock_err:
    cputs("Error: Cannot open connection.\n\r");
    return 1;
}



// --------------------------------------------------------------------------
// These routines do the actual talking:
// initscr() - Initialize screen before talk session
// local() - Switches to the `local' window
// remote() - Switches to the remote window
// cleanup() - Clean up after talk session
// talk() - This function does the actual talking
// --------------------------------------------------------------------------

static int xl,yl, xr,yr, where;

static void cleanup(int i)

{
    window(1,1,80,25);
    gotoxy(1,24);
    textattr(A_NORM);
    gotoxy(1,25);
    clreol();
    gotoxy(1,24);
    if (i == 1) cputs("You closed.");
    else if ( i == 2) cputs("Connection closed by remote user.");
    clreol();
}

static void remote(void)

{
    switch (scrnmode)
    {
	case 0:
	    break;
	case 1:
	    if (where != 1)
	    {
		yl = wherey(); xl = wherex();
		window(1,14,80,25);
		gotoxy(xr,yr);
		where = 1;
	    }
	    break;
    }
    textattr(a_remote);
}

static void local(void)

{
    switch (scrnmode)
    {
	case 0:
	    break;
	case 1:
	    if (where != 0)
	    {
		yr = wherey(); xr = wherex();
		window(1,1,80,12);
		gotoxy(xl,yl);
		where = 0;
	    }
	    break;
    }
    textattr(a_local);
}


static void init_screen(void)

{
    int staty;

    switch (scrnmode)
    {
	case 0:
	    textattr(A_NORM);
	    clrscr();
	    staty = 25;
	    break;
	case 1:
	    window(1,1,80,12);
	    textattr(a_local);
	    clrscr();
	    window(1,14,80,25);
	    textattr(a_remote);
	    clrscr();
	    staty = 13;
	    xr = yr = xl = yl = 1;
	    where = 0;
	    break;
    }

    window(1,1,80,25);
    gotoxy(1,staty);
    textattr(a_status);
    clreol();
    textattr(a_statusl);
    cprintf("\x18 Local: %s",localuser);
    textattr(a_status);
    cputs("    ");
    textattr(a_statusr);
    cprintf("\x19 Remote: %s@%s",remoteuser,remotehost);
    gotoxy(68,wherey());
    textattr(a_status);
    cputs("ESC - Exit");
    if (scrnmode == 1) window(1,1,80,12);
    else window(1,1,80,24);
    gotoxy(1,1);
    sound(1800);
    delay(100);
    nosound();
}


static int talk(void)

{
    char buf[100];
    char ch;
    int i;

    init_screen();

    while (tcp_tick(&ds))
    {
	if (kbhit())
	{
	    if ((ch = getch()) == 0x1b) return 1;
	    if ((unsigned) ch < ' ' && ch != '\r') continue;
	    local();
	    if (ch == '\r')
	    {
		ch = '\n';
		putch('\r');
	    }
	    putch(ch);
	    sock_putc(&ds,ch);
	}

	i = sock_fastread(&ds,buf,sizeof(buf));
	if (i > 0)
	{
	    int k;
	    remote();
	    for (k = 0; k < i; ++k)
	    {
		putch(buf[k]);
		if (buf[k] == '\n') putch('\r');
	    }
	}
    }
    return 2;
}

// --------------------------------------------------------------------------
// main()
// --------------------------------------------------------------------------

void main(int argc, char *argv[])

{
    int i = 0;

    if (init(argc,argv))
	exit(1);
    if (argc == 1)
	i = wait_invite();
    else
	i = invite();

    if (!i)
    {
	i = talk();
	sock_close(&ds);
	cleanup(i);
    }
    exit(0);
}
